#!/usr/bin/env python
"""
Read THEMIS GBO ASI data
"""
import logging
import warnings
from pathlib import Path
from datetime import datetime, timedelta
import xarray
from typing import Tuple, Sequence, Union, Optional, List
import numpy as np
from dateutil.parser import parse
import cdflib
import scipy.io
try:
    import h5py
except ImportError:
    h5py = None
try:
    from netCDF4 import Dataset
except ImportError:
    Dataset = None

Epoch = cdflib.cdfepoch()
cdfread = cdflib.cdfread.CDF


def load(path: Path,
         site: str = None,
         treq: datetime = None,
         calfn: Path = None) -> xarray.Dataset:
    """read THEMIS ASI camera data"""
# %% time slice (assumes monotonically increasing time)
    treq = _timereq(treq)

    imgs = _timeslice(path, site, treq)
# %% optional load calibration (az, el)
    data = xarray.Dataset({'imgs': imgs})
    data.attrs = imgs.attrs

    cal = None
    if calfn:
        cal = loadcal(calfn, site, treq)
    else:
        try:
            cal = loadcal(path, site, treq)
        except FileNotFoundError:
            pass

    if cal is not None:
        if cal.site is not None and cal.site != imgs.site:
            raise ValueError(f'cal site {cal.site} and data site {imgs.site} do not match. Was wrong calibration file used?')

        data = xarray.merge((data, cal))
        data.attrs = cal.attrs
        data.attrs.update(imgs.attrs)
        if data.caltime is not None:
            if (np.datetime64(data.caltime) >= data.time).any():
                raise ValueError('calibration is taken AFTER the images--may be incorrect lat/lon az/el plate scale')

    return data


def filetimes(fn: Path) -> List[datetime]:
    "prints the times available in a THEMIS ASI CDF file"""
    h = cdfread(fn)

    site = h.attget('Descriptor', 0)['Data'][:4].lower()

    return Epoch.to_datetime(h[f'thg_asf_{site}_epoch'][:])


def _timeslice(path: Path, site: str = None,
               treq: Union[datetime, Sequence[datetime], np.ndarray] = None) -> xarray.DataArray:
    """
    loads time slice of data
    """
# %% open CDF file handle (no close method)
    site, fn = _sitefn(path, site, treq)

    h = cdfread(fn)
# %% load image times
    try:
        time = Epoch.to_datetime(h[f'thg_asf_{site}_epoch'][:], to_np=True)
    except KeyError:
        time = np.array(list(map(datetime.utcfromtimestamp,
                                 h[f'thg_asf_{site}_time'][:])))
# %% time request handling
    if treq is None:
        i = slice(None)
    else:
        treq = np.atleast_1d(treq)

        if treq.size == 1:
            # Note: arbitrarily allowing up to 1 second time offset from request
            if (treq < time-timedelta(seconds=1)).all() | (treq > time+timedelta(seconds=1)).all():
                raise ValueError(f'requested time outside {fn}')

            i = abs(time - treq[0]).argmin()
        elif treq.size == 2:  # start, end
            i = (time >= treq[0]) & (time <= treq[1])
        else:
            raise ValueError('for now, time req is single time or time range')

    imgs = h[f'thg_asf_{site}'][i]
    if imgs.ndim == 2:
        imgs = imgs[None, ...]

    time = time[i]
    if isinstance(time, datetime):
        time = [time]
    elif len(time) == 0:
        raise ValueError(f'no times were found with requested time bounds {treq}')

    return xarray.DataArray(imgs, {'time': time}, ['time', 'y', 'x'],
                            attrs={'filename': fn.name, 'site': site})


def _sitefn(path: Path, site: str = None,
            treq: Union[datetime, Sequence[datetime], np.ndarray] = None) -> Tuple[str, Path]:
    """
    gets site name and CDF key from filename (!)
    """

    path = Path(path).expanduser()

    if path.is_dir():
        if not isinstance(site, str):
            raise ValueError('Must specify filename OR path and site and time')

        # FIXME: assumes time bounds don't cross file boundaries
        if treq is None:
            raise ValueError('Must specify filename OR path and site and time')
        elif isinstance(treq, datetime):
            t0 = treq
        elif isinstance(treq[0], datetime) and len(treq) in (1, 2):
            t0 = treq[0]
        else:
            raise ValueError('Must specify filename OR path and site and time')

        fn = path / f'thg_l1_asf_{site}_{t0.year}{t0.month:02d}{t0.day:02d}{t0.hour:02d}_v01.cdf'
        if not fn.is_file():
            # try to use last time in file, if first time wasn't covered
            if isinstance(treq, datetime):
                raise FileNotFoundError(fn)

            t0 = treq[-1]

            fn = path / f'thg_l1_asf_{site}_{t0.year}{t0.month:02d}{t0.day:02d}{t0.hour:02d}_v01.cdf'

    elif path.is_file():
        if site:
            raise ValueError('specify filename OR site')

        fn = path

        h = cdfread(fn)
        site = h.attget('Descriptor', 0)['Data'][:4].lower()
    else:
        raise FileNotFoundError(f'video file not found in {path}')

    assert isinstance(site, str)

    return site, fn


def _timereq(treq: datetime = None) -> Optional[datetime]:

    if treq is None or isinstance(treq, datetime):
        pass
    elif isinstance(treq, str):
        treq = parse(treq)
    elif isinstance(treq[0], str):  # type: ignore
        treq = list(map(parse, treq))  # type: ignore
    elif isinstance(treq[0], datetime):  # type: ignore
        pass
    else:
        raise TypeError(f'not sure what treq {treq} is')

    return treq


def _downsample(imgs: xarray.Dataset, az: np.ndarray, el: np.ndarray,
                x: np.ndarray, y: np.ndarray) -> xarray.Dataset:
    """downsamples cal data to match image data
    because of the discontinuous nature of the calibration data, typical resampling is not valid.
    Figured better to add a little error with plain decimation rather than enormous error with invalid technique.
    """
    if az.shape == imgs.shape[1:]:
        return az, el, x, y

    downscale = (az.shape[0] // imgs.shape[1],
                 az.shape[1] // imgs.shape[2])

    logging.warning(f'downsizing calibration az/el data by factors of {downscale} to match image data')

    az = az[::downscale[0], ::downscale[1]]
    el = el[::downscale[0], ::downscale[1]]
    x = np.arange(az.shape[1])
    y = np.arange(az.shape[0])

    return az, el, x, y


def loadcal_file(fn: Path) -> xarray.Dataset:
    """
    reads data mapping themis gbo asi pixels to azimuth,elevation
    calibration data url is
    http://data.phys.ucalgary.ca/sort_by_project/THEMIS/asi/skymaps/new_style/
    """
    site = None
    time = None
    fn = Path(fn).expanduser()
    if not fn.is_file():
        raise FileNotFoundError(f'calibration file not found: {fn}')

    if fn.suffix == '.cdf':
        site = fn.name.split('_')[3]
        h = cdfread(fn)
        az = h[f'thg_asf_{site}_azim'][0]
        el = h[f'thg_asf_{site}_elev'][0]
        lat = h[f'thg_asc_{site}_glat']
        lon = (h[f'thg_asc_{site}_glon'] + 180) % 360 - 180  # [0,360] -> [-180,180]
        alt_m = h[f'thg_asc_{site}_alti']
        x = y = h[f'thg_asf_{site}_c256']
        time = datetime.utcfromtimestamp(h[f'thg_asf_{site}_time'][-1])
    elif fn.suffix == '.sav':
        site = fn.name.split('_')[2]
        # THEMIS SAV calibration files written with glitch from bug in IDL
        warnings.simplefilter("ignore", UserWarning)
        h = scipy.io.readsav(fn, python_dict=True, verbose=False)
        warnings.resetwarnings()

        az = h['skymap']['full_azimuth'][0]
        el = h['skymap']['full_elevation'][0]
        lat = h['skymap']['site_map_latitude'].item()
        lon = (h['skymap']['site_map_longitude'].item() + 180) % 360 - 180  # [0,360] -> [-180,180]
        alt_m = h['skymap']['site_map_altitude'].item()
        x = h['skymap']['full_column'][0][0, :]
        y = h['skymap']['full_row'][0][:, 0]
        try:
            tstr = h['skymap']['generation_info'][0][0][2]
            time = datetime(int(tstr[:4]), int(tstr[4:6]), int(tstr[6:8]), int(tstr[8:10]))
        except (KeyError, ValueError):
            if h['skymap']['site_unix_time'] > 0:
                tutc = h['skymap']['site_unix_time']
            elif h['skymap']['imager_unix_time'] > 0:
                tutc = h['skymap']['imager_unix_time']
            else:
                tutc = None

            if tutc is not None:
                time = datetime.utcfromtimestamp(tutc)
            else:  # last resort
                time = datetime(int(fn.name[19:23]),
                                int(fn.name[23:25]),
                                int(fn.name[25:27]))

    elif fn.suffix == '.h5':
        if h5py is None:
            raise ImportError('pip install h5py')

        with h5py.File(fn, 'r') as h:
            az = h['az'][:]
            el = h['el'][:]
            lat = h['lla'][0]
            lon = h['lla'][1]
            alt_m = h['lla'][2]
            x = h['x'][0, :]
            y = h['y'][:, 0]
    elif fn.suffix == '.nc':
        if Dataset is None:
            raise ImportError('pip install netCDF4')

        with Dataset(fn, 'r') as h:
            az = h['az'][:]
            el = h['el'][:]
            lat = h['lla'][0]
            lon = h['lla'][1]
            alt_m = h['lla'][2]
            x = h['x'][0, :].astype(int)
            y = np.flipud(h['y'][:, 0]).astype(int)
    else:
        raise ValueError(f'{fn} calibration file format is not known to this program.')

    cal = xarray.Dataset({'az': (('y', 'x'), az),
                          'el': (('y', 'x'), el)},
                         coords={'y': y, 'x': x},
                         attrs={'lat': lat, 'lon': lon, 'alt_m': alt_m,
                                'site': site, 'calfilename': fn.name,
                                'caltime': time})

    return cal


def loadcal(path: Path, site: str = None, time: datetime = None) -> xarray.Dataset:
    path = Path(path).expanduser()

    if path.is_file() and site is None or time is None:
        try:
            return loadcal_file(path)
        except KeyError:
            return None
    elif path.is_file():
        path = path.parent

    assert isinstance(site, str)
    fn = _findcal(path, site, time)

    return loadcal_file(fn)


def _findcal(path: Path, site: str, time: datetime) -> Path:
    """
    attempt to find nearest previous time calibration file
    """

    if not path.is_dir():
        raise FileNotFoundError(str(path))

    if not isinstance(site, str) or len(site) != 4:
        raise ValueError(f'site code is four characters e.g. fykn.  You gave:   {site}')

    if isinstance(time, str):
        time = parse(time)

    if isinstance(time, (list, tuple, np.ndarray)):
        time = time[0]  # assume first time is earliest

    if not isinstance(time, datetime):
        raise TypeError(f'must specify single datetime, you gave:  {time}')
# %% CDF .cdf
    fcdf = list(path.glob(f'thg_l2_asc_{site}_*.cdf'))
    cdates = [loadcal(fn).caltime for fn in fcdf]

    datecdf = None
    if cdates:
        for i, date in enumerate(cdates):
            if date < time:
                break
        if date < time:
            datecdf = date
            icdf = len(cdates) - (i+1)
# %% IDL .sav
    fsav = list(path.glob(f'themis_skymap_{site}_*.sav'))
    sdates = [loadcal(fn).caltime for fn in fsav]

    datesav = None
    if sdates:
        for i, date in enumerate(sdates):
            if date < time:
                break
        if date < time:
            datesav = date
            isav = len(sdates) - (i+1)

# %% get result
    if not sdates and not cdates:
        raise FileNotFoundError(f'could not find cal file for {site} {time}  in {path}')
    elif datecdf is None:
        return fsav[isav]
    elif datesav is None:
        return fcdf[icdf]
# tiebreaker
    diff = [abs(datecdf-time), abs(datesav-time)]
    idff = diff.index(min(diff))

    if idff == 0:
        return fcdf[icdf]
    else:
        return fsav[isav]



from dateutil.parser import parse
from datetime import datetime, timedelta
from pathlib import Path
import requests
import numpy as np
import warnings
from typing import Sequence, Union
from .io import load, loadcal, filetimes  # noqa: F401

URL = 'http://themis.ssl.berkeley.edu/data/themis/thg/l1/asi/'
CALURL = 'http://themis.ssl.berkeley.edu/data/themis/thg/l2/asi/cal/'


def urlretrieve(url: str, fn: Path, overwrite: bool = False):
    if not overwrite and fn.is_file() and fn.stat().st_size > 10000:
        print(f'SKIPPED {fn}')
        return
# %% prepare to download
    R = requests.head(url, allow_redirects=True, timeout=10)
    if R.status_code != 200:
        warnings.warn(f'{url} not found. \n HTTP ERROR {R.status_code}')
        return
# %% download
    print(f'downloading {int(R.headers["Content-Length"])//1000000} MBytes:  {fn.name}')
    R = requests.get(url, allow_redirects=True, timeout=10)
    with fn.open('wb') as f:
        f.write(R.content)


def download(treq: Sequence[Union[str, datetime]],
             site: Sequence[str], odir: Path,
             overwrite: bool = False, host: str = URL):
    """
    treq: time request e.g. '2012-02-01T03' or a time range ['2012-02-01T03', '2012-02-01T05']
    site: camera site e.g. gako, fykn
    odir: output directory e.g. ~/data
    overwrite: overwrite existing files--normally wasteful of download space, unless you had a corrupted download.
    host: website hosting files
    """
    if not host:
        raise ValueError(f'Must specify download host, e.g. {URL}')

# %% sanity checks
    odir = Path(odir).expanduser().resolve()
    odir.mkdir(parents=True, exist_ok=True)

    if isinstance(site, str):
        site = [site]

    if isinstance(treq, (str, datetime)):
        treq = [treq]

    start = parse(treq[0]) if isinstance(treq[0], str) else treq[0]
    if len(treq) == 2:
        end = parse(treq[1]) if isinstance(treq[1], str) else treq[1]
    else:
        end = start

    if end < start:
        raise ValueError('start time must be before end time!')
# %% start download
    for s in site:
        download_cal(s, odir, overwrite)
        for T in np.arange(start, end+timedelta(hours=1), timedelta(hours=1)):
            t = T.astype(datetime)
            fpath = (f'{host}{s}/{t.year:4d}/{t.month:02d}/'
                     f'thg_l1_asf_{s}_{t.year:4d}{t.month:02d}{t.day:02d}{t.hour:02d}_v01.cdf')

            urlretrieve(fpath, odir / fpath.split('/')[-1])


def download_cal(site: str, odir: Path, owerwrite: bool = False, host: str = CALURL):
    fpath = f"{host}thg_l2_asc_{site}_19700101_v01.cdf"
    urlretrieve(fpath,  odir / fpath.split('/')[-1])



#!/usr/bin/env python
"""
paint HiST field of view onto Themis

"""
from datetime import datetime
from matplotlib.pyplot import show
from themisasi import altfiducial
from themisasi.plots import plotthemis
try:
    import seaborn as sns
    sns.set_context('talk')
except ImportError:
    pass


sitereq = 'whit'  # 'fykn' 'gako' 'mcgr'
# for this date, kian and whit are binned 32x32, we'd need to figure out the non-trivial binning to use them.
# kian and whit are far away, so this is a low priority.

# treq = (datetime(2013,4,14,8,30),datetime(2013,4,14,8,32)) #when big auroral hook  form for orientation happens
treq = (datetime(2013, 4, 14, 8, 53), datetime(
    2013, 4, 14, 8, 56))  # HIST event
odir = None

moviefn = f'THEMIS/{sitereq}_hst0,1_{treq[0].strftime("%Y-%m-%dT%H%M")}.mkv'
# %%

if sitereq == 'fykn':
    asifn = 'THEMIS/thg_l1_asf_fykn_2013041408_v01.cdf'
    asical = 'THEMIS/cal/themis_skymap_fykn_20110305-+_vXX.sav'
elif sitereq == 'gako':
    asifn = 'THEMIS/thg_l1_asf_gako_2013041408_v01.cdf'
    asical = 'THEMIS/cal/themis_skymap_gako_20110305-+_vXX.sav'
elif sitereq == 'mcgr':
    asifn = 'THEMIS/thg_l1_asf_mcgr_2013041408_v01.cdf'
    asical = 'THEMIS/cal/themis_skymap_mcgr_20110305-+_vXX.sav'
elif sitereq == 'inuv':
    asifn = 'THEMIS/thg_l1_asf_inuv_2013041408_v01.cdf'
    asical = 'THEMIS/cal/themis_skymap_inuv_20150401-+_vXX.sav'
elif sitereq == 'kian':
    asifn = 'THEMIS/thg_l1_ast_kian_20130414_v01.cdf'
    asical = 'THEMIS/cal/themis_skymap_kian_20111107-+_vXX.sav'
elif sitereq == 'whit':
    asifn = 'THEMIS/thg_l1_ast_whit_20130414_v01.cdf'
    asical = 'THEMIS/cal/themis_skymap_whit_20150814-+_vXX.sav'
else:
    raise ValueError(f'site {sitereq} not configured')

ocal = ['~/code/histfeas/precompute/hst0cal.h5',
        '~/code/histfeas/precompute/hst1cal.h5']

# %%
try:
    plotthemis(imgs, t, site, treq, moviefn, rows=R, cols=C,
               ext=(wC[0, 0], wC[0, -1], wR[-1, 0], wR[0, 0]))
except NameError:
    # 60 second computation
    imgs, R, C, t, site, wR, wC = altfiducial(asifn, asical, ocal, treq, odir)


show()



#!/usr/bin/env python
"""
UAF DASC with THEMIS GBO ASI FOV: Take a slice

The API to take the slice is very old code with a cumbersome, awkward API.
I have started by making a monkey-patched use of this API.
If the project proceeds, I would perhaps redo the API much more cleanly.

./DascThemisFOV.py ~/data/themis/themis_skymap_gako_20070401.sav ~/data/dasc/PKR_DASC_0000_20080326_070857.000.FITS \
     ../dascutils/cal/PKR_DASC_20110112_AZ_10deg.fits ../dascutils/cal/PKR_DASC_20110112_EL_10deg.fits

DASC coord: 65.1260, -147.479
DASC exposure: 1 sec.
"""
from argparse import ArgumentParser
import numpy as np
import themisasi.fov as taf
import themisasi as ta
import themisasi.plots as tap
import dascutils.io as dio
try:
    from matplotlib.pyplot import show
except ImportError:
    show = None


def main():
    p = ArgumentParser(description='register/plot other camera FOV onto Themis FOV')
    p.add_argument('themiscal', help='ASI calibration')
    p.add_argument('dasccal', help='DASC cal', nargs=3)
    p.add_argument('-o', '--ofn', help='write output plot')
    p.add_argument('-v', '--verbose', action='store_true')
    P = p.parse_args()

# %% load az/el data
    themis = ta.loadcal(P.themiscal)
    if themis.site == 'gako':
        themis.attrs['Baz'] = 21.028
        themis.attrs['Bel'] = 75.816
        themis.attrs['Bepoch'] = '2010-01-01'
        themis.attrs['verbose'] = False
        themis.attrs['srpts'] = np.logspace(5, 6.9, 40)
        themis['cutcol'] = np.arange(50, themis.x.size - 75, 1, dtype=int)

    dasc = dio.load(*P.dasccal)
    if 'lat' not in dasc.attrs:
        dasc.attrs['lat'] = 65.126
        dasc.attrs['lon'] = -147.479
        dasc.attrs['alt_m'] = 200.
        dasc.attrs['Baz'] = 21.145  # Baz = declination
        dasc.attrs['Bel'] = 77.464  # Bel = inclination
        dasc.attrs['Bepoch'] = '2010-01-01'
        dasc.attrs['verbose'] = False
        # np.linspace(10**4.6,10**6.8,40) #np.logspace(4.6, 6.4, 40)
        dasc.attrs['srpts'] = np.logspace(4.6, 7.5, 40)
        dasc['cutcol'] = np.arange(75, dasc.x.size - 100, 1, dtype=int)
# %% extract common 2-D slice (plane)
    # paint HiST field of view onto Themis
    themis, dasc = taf.mergefov(themis, dasc, projalt=110e3, method='MZslice')
    themis = taf.line2plane(themis)
    dasc = taf.line2plane(dasc)
    if show is None:
        return
# %% plot slice pixel mask
    tap.jointazel(
        themis, P.ofn, 'Themis Gakona slice toward DASC and magnetic zenith')
    tap.jointazel(
        dasc, P.ofn, 'DASC Poker Flat slice toward Themis Gakona and magnetic zenith')

    show()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
UAF DASC FOV overlaid on THEMIS GBO ASI FOV

./DascThemisFOV.py ~/data/themis/themis_skymap_gako_20070401.sav ../dascutils/tests/PKR_DASC_0428_20151007_082305.930.FITS \
       ../dascutils/cal/PKR_DASC_20110112_AZ_10deg.fits ../dascutils/cal/PKR_DASC_20110112_EL_10deg.fits
"""
from argparse import ArgumentParser
import themisasi.plots as tap
import themisasi.fov as taf
import themisasi as ta
import dascutils.io as dio
from matplotlib.pyplot import show


def main():
    p = ArgumentParser(description='register/plot other camera FOV onto Themis FOV')
    p.add_argument('themiscal', help='ASI calibration')
    p.add_argument('dasccal', help='DASC cal', nargs=3)
    p.add_argument('-o', '--ofn', help='write output plot')
    P = p.parse_args()

# %% load data
    themis = ta.loadcal(P.themiscal)
    dasc = dio.load(*P.dasccal)
# %% merge FOV
    # paint HiST field of view onto Themis
    themis, dasc = taf.mergefov(themis, dasc, projalt=110e3, method='perimeter')
# %% plot joint az/el contours
    tap.jointazel(themis, P.ofn, 'Themis Gakona overlaid on Poker Flat ASI')

    show()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Playback THEMIS ASI videos
* optionally, save images to stack of PNGs

PlotThemis ~/data/themis/thg_l1_asf_fykn_2013041408_v01.cdf
"""
from argparse import ArgumentParser
import themisasi as ta
import themisasi.plots as tap


def main():
    p = ArgumentParser(description=' reads THEMIS GBO ASI CDF files and plays high speed video')
    p.add_argument('path', help='directory THEMIS ASI video files are in')
    p.add_argument('site', help='THEMIS ASI site code e.g. fykn')
    p.add_argument('treq', help='time or start,stop time range requested', nargs='+')
    p.add_argument('-o', '--odir', help='write video to this directory')
    P = p.parse_args()

    imgs = ta.load(P.path, site=P.site, treq=P.treq)
# %% plot
    tap.plotazel(imgs)

    tap.plotasi(imgs, P.odir)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
example of image projection to an altitude.
This assumes all the brightness comes from a thin layer at that altitude, which
is a common first-order approximation made in multi-instrument studies, for example using
auroral video with GNSS TEC measurements.
"""
from pathlib import Path
import themisasi as ta
import themisasi.projections as tap


datadir = Path('~/data/2016-11-29').expanduser()

time_query = '2016-11-29T12'

ta.download(time_query, 'mcgr', datadir)

dat = ta.load(datadir, site='mcgr', treq=time_query)

tap.asi_projection(dat, 110e3)
